# Used by random_variable.py.
DIDNT_PASS_EITHER_RV_OR_MODEL = (
    "You didn't pass either a pymc random variable or pymc.Model object into the "
    "RandomVariable constructor."
)

DIDNT_PASS_PYMC_MODEL_OBJ = (
    "You didn't pass a pymc.Model object into the RandomVariable constructor."
)

CANT_FIND_PYMC_RV_IN_PYMC_MODEL_OBJ = (
    "The pymc random variable name that you passed into the RandomVariable "
    "constructor didn't match the name of any of the pymc variables attached to "
    "the pymc.Model that you also passed into the RandomVariable constructor."
)

BAD_RV_DIMS = (
    "The pymc random variable that is attached to the pymc.Model object that you passed into the "
    "RandomVariable constructor had more than two dimensions (i.e. you wanted to create a RandomVariable "
    "that is neither a scalar nor a matrix)."
)

UNSUPPORTED_PYMC_RV = (
    "It appears that pymc.Model object that you passed into the RandomVariable constructor "
    "contains a pymc random variable object that is unsupported."
)

UNSUPPORTED_RV_TYPE = "You asked RandomVariableFactory to create a RandomVariable that it doesn't know how to create."

BAD_VAL_MAP = (
    "You passed into RandomVariableFactory a vals list that contains items that are neither int, float, "
    "nor numpy.ndarray."
)

UNSUPPORTED_CONSTR = (
    "You created a Problem object that you passed into partial_optimize that contained a constraint "
    "that is unsupported right now."
)

# Used by partial_optimize.py.
CANT_DETERMINE_CURVATURE = (
    "You called partial_optimize with a Problem object that contains neither a Minimize nor a "
    "Maximize statement; this is not supported."
)

# Used by prob.py.
UNSUPPORTED_BOUND_TYPE = (
    "You called prob with an upper bound type that is not supported."
)



from cvxpy import Variable
from cvxpy.expressions.cvxtypes import add_expr, neg_expr
from cvxpy.atoms import maximum, square, inv_pos
from cvxpy.lin_ops import lin_utils
from cvxpy.constraints import *

from cvxstoc.expectation import expectation
from cvxstoc.expectation import clamp_or_sample_rvs
from cvxstoc.utils import replace_rand_vars
import cvxstoc.strings

import math
import copy


class Phi:
    NONE = 1
    HINGE = 2


class prob:
    """
    Notes:
    - constr must be in the form Prob( f(x) >= g(x) ), where f is convex and g is concave (g is often just 0).
    - if phi == Phi.HINGE, then we expect Prob( f(x) >= g(x) ) to be <= beta, where beta is usually small
      (e.g., beta = 0.05).
    """

    def __init__(self, constr, num_samples, phi=Phi.HINGE):
        self.constr = constr
        self.num_samples = num_samples
        self.phi = phi

        if self.phi == Phi.HINGE:
            self.alpha = Variable(nonneg=True)

    def __le__(self, beta):
        constr_rearranged = (
            self.constr.args[1] - self.constr.args[0]
        )  # Note: constr.rh_exp <==> f(x)
        # and constr.lh_exp <==> g(x)
        # in the "Note" above
        self.conservative = expectation(
            maximum(0.0, constr_rearranged + self.alpha), self.num_samples
        )

        if self.phi == Phi.HINGE:
            return self.conservative - self.alpha * beta <= 0
        else:
            raise Exception(strings.UNSUPPORTED_BOUND_TYPE)

    @property
    def value(self):
        constr_rearranged = self.constr.args[0] - self.constr.args[1]

        true_cnt = 0.0
        rvs2samples, ignore = clamp_or_sample_rvs(
            constr_rearranged,
            rvs2samples={},
            want_de=False,
            want_mf=False,
            num_samples=self.num_samples,
            num_burnin_samples=0,
        )
        for s in range(self.num_samples):
            constr_rearranged_copy = replace_rand_vars(constr_rearranged)
            constr_rearranged_copy_rlzd = clamp_or_sample_rvs(
                constr_rearranged_copy,
                rvs2samples,
                want_de=False,
                want_mf=False,
                sample_idx=s,
            )
            if constr_rearranged_copy_rlzd.value <= 0:
                true_cnt += 1.0

        return true_cnt / self.num_samples



import sys  # TODO: temp

sys.path.insert(1, "/Users/alnurali/shared_code/cvxpy/cvxpy")  # TODO: temp
import cvxpy as cvx
import cvxpy.stochastic
from cvxpy import CVXOPT

import numpy as np
import math
import scipy.stats

import matplotlib.pyplot as plt

# from matplotlib.backends.backend_pdf import PdfPages
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm


class HowEstUnconstrOptVal:
    AVERAGE_SAAS = 1
    MEANFIELD = 2


class HowEstConstrOptVal:
    LAGRANGIAN = 1


class ConfInt:
    T = 1
    BOOT_TMP = 2


class HowGetCritVal:
    T = 1
    NORMAL = 2


class ConstrDiagnostics:
    def est_optval(
        self, how, obj, constrs
    ):  # Note: obj and constrs are expected to be *deterministic* functions

        if how == HowEstConstrOptVal.LAGRANGIAN:
            lagrangian = obj
            for i in range(len(constrs)):
                stoch_constr = constrs[i].lh_exp - constrs[i].rh_exp
                (m, n) = stoch_constr.size
                dual_var = 0.1 * np.ones(m)  # 0.1+np.random.rand(m)
                lagrangian += stoch_constr.T * dual_var
            prob = cvx.Problem(cvx.Minimize(lagrangian))
            prob.solve()
            return prob.value


class UnconstrDiagnostics:
    def est_optgap(self, how, stoch_objf, num_replications, num_samples, alpha):
        (ignore, fhatmean, fhatvar, ignore, ignore) = self.est_cost(
            how, stoch_objf, num_replications, alpha
        )
        (pstarhatmean, pstarhatvar, ignore, ignore) = self.est_optval(
            how, stoch_objf, num_replications, num_samples, alpha
        )
        return self.est_optgap_core(fhatmean, fhatvar, pstarhatmean, pstarhatvar, alpha)

    def est_optgap_core(self, fhatmean, fhatvar, pstarhatmean, pstarhatvar, alpha):
        gaphat = fhatmean - pstarhatmean
        pm = scipy.stats.norm.ppf(1.0 - (1.0 - alpha) / 2.0) * np.sqrt(
            fhatvar + pstarhatvar
        )
        return (gaphat, gaphat + pm, gaphat - pm)

    def est_cost(self, stoch_objf, num_replications, alpha):
        # Evaluate the SAA objective many times; compute the mean and variance
        fhats = [
            cvx.stochastic.expectation(stoch_objf, 1).value
            for i in range(num_replications)
        ]
        fhatmean = np.mean(fhats)

        # Compute confidence intervals
        fhatvar = np.var(fhats, ddof=1)

        fhatstderr = np.sqrt(1.0 / num_replications * fhatvar)
        fhatucb = fhatmean + UnconstrDiagnostics.get_critval(alpha) * fhatstderr
        fhatlcb = fhatmean - UnconstrDiagnostics.get_critval(alpha) * fhatstderr

        # Return
        return (fhatmean, fhatvar, fhatucb, fhatlcb)

    def est_optval(
        self,
        how_est_optval,
        stoch_objf,
        num_replications,
        num_samples,
        how_conf_int,
        alpha,
        A=None,
        b=None,
    ):

        if how_est_optval == HowEstUnconstrOptVal.AVERAGE_SAAS:

            if how_conf_int == ConfInt.T:

                pstarhats = []
                critval = UnconstrDiagnostics.get_critval(
                    alpha, HowGetCritVal.T, num_replications - 1
                )
                for i in range(num_replications):
                    prob = cvx.Problem(
                        cvx.Minimize(
                            cvx.stochastic.expectation(stoch_objf, num_samples)
                        )
                    )
                    prob.solve()
                    pstarhats.append(prob.value)

                pstarhatmean = np.mean(pstarhats)

                pstarhatvar = np.var(pstarhats, ddof=1)

                pstarhatstderr = np.sqrt(1.0 / num_replications * pstarhatvar)
                pstarhatucb = pstarhatmean + critval * pstarhatstderr
                pstarhatlcb = pstarhatmean - critval * pstarhatstderr

            else:  # Use bootstrap

                dims = A.shape
                (m, n) = dims.rows, dims.cols
                x = cvx.Variable(n)

                A_samples = A.sample(num_samples)
                b_samples = b.sample(num_samples)

                objf = 0
                for idx in range(num_samples):
                    objf += cvx.sum_squares(
                        A_samples[idx, :, :] * x - b_samples[idx, :]
                    )
                prob = cvx.Problem(cvx.Minimize((1.0 / num_samples) * objf))
                prob.solve()
                pstarhatmean = prob.value

                pstarhats = []
                for i in range(
                    num_replications
                ):  # Each iteration here is a bootstrap replication
                    idxes = np.random.randint(0, num_samples, num_samples)

                    objf = 0
                    for idx in idxes:
                        objf += cvx.sum_squares(
                            A_samples[idx, :, :] * x - b_samples[idx, :]
                        )

                    prob = cvx.Problem(cvx.Minimize((1.0 / num_samples) * objf))
                    prob.solve()

                    pstarhats.append(prob.value)

                pstarhatvar = np.var(pstarhats)

                pstarhatucb = np.percentile(pstarhats, 100.0 * (1.0 - alpha / 2.0))
                pstarhatlcb = np.percentile(pstarhats, 100.0 * (alpha / 2.0))

        else:  # Use mean field approximation (no conf ints)
            prob = cvx.Problem(
                cvx.Minimize(cvx.stochastic.expectation(stoch_objf, want_mf=True))
            )
            prob.solve()
            pstarhatmean = prob.value

            pstarhatvar, pstarhatucb, pstarhatlcb = None, None, None

        # Return
        return (pstarhatmean, pstarhatvar, pstarhatucb, pstarhatlcb)

    @staticmethod
    def get_critval(
        alpha, how=HowGetCritVal.NORMAL, df=None
    ):  # Note: alpha is (expected to be) small here
        if how == HowGetCritVal.T:
            return scipy.stats.t.ppf(1.0 - alpha / 2.0, df)
        else:  # Use inverse normal cdf
            return scipy.stats.norm.ppf(1.0 - alpha / 2.0)


class TestDiagnostics:
    def __init__(self):
        np.random.seed(1)

        self.unconstr_diag = UnconstrDiagnostics()
        self.constr_diag = ConstrDiagnostics()

        self.fontsize = 24  # 16 For plots
        self.fontsize_axis = 16
        self.linewidth = 3  # 2

    def test_least_squares_plot_saa_quality(
        self, A, b, P, q, r, num_samples_list, fn, num_replications, alpha
    ):
        dims = A.shape
        (m, n) = dims.rows, dims.cols

        # Solve the SAA
        pstarhat_avgs, pstarhat_ucbs, pstarhat_lcbs = [], [], []
        critval = UnconstrDiagnostics.get_critval(
            alpha, HowGetCritVal.T, num_replications - 1
        )
        for num_samples in num_samples_list:
            print "Solving a SAA to the stochastic least squares problem with [%d] Monte Carlo samples." % num_samples
            pstarhats = []
            for i in range(num_replications):
                x = cvx.Variable(n)
                prob = cvx.Problem(
                    cvx.Minimize(
                        cvx.stochastic.expectation(
                            cvx.sum_squares(A * x - b), num_samples
                        )
                    )
                )
                prob.solve()
                # print "Got objective value [%f]." % prob.value
                pstarhats.append(prob.value)

            pstarhat_avg = np.mean(pstarhats)
            pstarhat_avgs.append(pstarhat_avg)

            pstarhat_se = scipy.stats.sem(pstarhats)
            pstarhat_ucbs.append(pstarhat_avg + critval * pstarhat_se)
            pstarhat_lcbs.append(pstarhat_avg - critval * pstarhat_se)

        # Solve the true problem
        prob = cvx.Problem(cvx.Minimize(cvx.quad_form(x, P) - 2 * x.T * q + r))
        print "Solving the true stochastic least squares problem."
        prob.solve()
        # print "Got objective value [%f]." % prob.value
        pstars = np.tile(prob.value, len(num_samples_list))

        # Make plots
        print "Plotting results to [" + fn + "] (on a linear/log scale)."
        fig, ax = plt.subplots()
        plt.rc("text", usetex=True)
        # plt.rc('font', family='serif')
        ax.plot(
            num_samples_list,
            pstarhat_avgs,
            "-",
            label=r"$\hat{p}^*_N$",
            color="blue",
            lw=self.linewidth,
        )
        ax.fill_between(
            num_samples_list,
            pstarhat_ucbs,
            pstarhat_lcbs,
            facecolor="lightskyblue",
            alpha=0.5,
        )
        ax.plot(
            num_samples_list,
            pstars,
            "-",
            label=r"$p^*$",
            color="crimson",
            lw=self.linewidth,
        )

        ax.legend(loc="lower right", fontsize=self.fontsize)
        ax.set_xlabel(
            r"\# Monte Carlo samples $N$", fontsize=self.fontsize
        )  # number of Monte Carlo samples $N$
        # ax.xticks(num_samples_list)
        # plt.savefig(fn, bbox_inches="tight")
        ax.set_xscale("log")
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontsize(self.fontsize_axis)
        fig.savefig(fn, bbox_inches="tight")
        print "All done."

    def test_least_squares_plot_cost_ests(
        self, A, b, P, q, r, x, num_replications_list, alpha, fn
    ):
        # Estimate the true objective value
        stoch_objf = cvx.sum_squares(A * x - b)
        fhatmeans, fhatvars, fhatucbs, fhatlcbs = [], [], [], []
        for num_replications in num_replications_list:
            print "Averaging [%d] times the obj. vals. at a candidate point of SAAs with 1 Monte Carlo sample to the stochastic least squares objective." % num_replications
            (fhatmean, fhatvar, fhatucb, fhatlcb) = self.unconstr_diag.est_cost(
                stoch_objf, num_replications, alpha
            )
            fhatmeans.append(fhatmean)
            fhatvars.append(fhatvar)
            fhatucbs.append(fhatucb)
            fhatlcbs.append(fhatlcb)

        # Compute the true objective value
        print "Evaluating the true stochastic least squares objective value at a candidate point."
        fstar = (cvx.quad_form(x, P) - 2 * x.T * q + r).value
        fstars = np.tile(fstar, len(num_replications_list))

        # Make plots
        print "Plotting results to [" + fn + "] (on a linear/log scale)."
        fig, ax = plt.subplots()
        plt.rc("text", usetex=True)
        ax.plot(
            num_replications_list,
            fhatmeans,
            "-",
            label=r"$\hat{f}_{0,N}(x)$",
            color="blue",
            lw=self.linewidth,
        )
        ax.fill_between(
            num_replications_list,
            fhatucbs,
            fhatlcbs,
            facecolor="lightskyblue",
            alpha=0.5,
        )
        ax.plot(
            num_replications_list,
            fstars,
            "-",
            label=r"$f_0(x)$",
            color="crimson",
            lw=self.linewidth,
        )

        ax.legend(loc="upper right", fontsize=self.fontsize)
        ax.set_xlabel(r"\# samples $N$", fontsize=self.fontsize)
        ax.set_xscale("log")
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontsize(self.fontsize_axis)
        fig.savefig(fn, bbox_inches="tight")
        print "All done."

        # Return
        return (fhatmeans, fhatvars, fstars)

    def test_least_squares_plot_optval_ests(
        self,
        A,
        b,
        P,
        q,
        r,
        num_replications_list,
        num_samples_list,
        alpha,
        fn1,
        fn2,
        fn3,
        fn4,
        fn5,
        fhatmeans,
        fstars,
    ):
        # Compute the true optimal value
        dims = A.shape
        m, n = dims.rows, dims.cols

        x = cvx.Variable(n)
        prob = cvx.Problem(cvx.Minimize(cvx.quad_form(x, P) - 2 * x.T * q + r))
        print "Solving the true stochastic least squares problem."
        prob.solve()
        pstar = prob.value

        # Compute the true optimality gap
        gap = fstars[0] - pstar

        # Estimate the optimality gap (using Monte Carlo)
        X = len(num_replications_list)
        Y = len(num_samples_list)

        pstarhats = np.zeros((X, Y))
        pstarhat_ucbs = np.zeros((X, Y))
        pstarhat_lcbs = np.zeros((X, Y))

        gaphats = np.zeros((X, Y))

        errors = np.zeros((X, Y))
        for i in range(X):  # range(X-1,-1,-1):
            num_replications = num_replications_list[i]

            for j in range(Y):  # range(Y-1,-1,-1):
                num_samples = num_samples_list[j]

                # Estimate the true optimal value
                stoch_objf = cvx.sum_squares(A * x - b)
                print "Averaging the opt. vals. of [%d] SAAs with [%d] Monte Carlo samples to the stoch. least squares problem; requesting t-based conf. ints." % (
                    num_replications,
                    num_samples,
                )
                (
                    pstarhatmean,
                    pstarhatvar_ignore,
                    pstarhatucb,
                    pstarhatlcb,
                ) = self.unconstr_diag.est_optval(
                    HowEstUnconstrOptVal.AVERAGE_SAAS,
                    stoch_objf,
                    num_replications,
                    num_samples,
                    ConfInt.T,
                    alpha,
                )
                pstarhats[i, j] = pstarhatmean
                pstarhat_ucbs[i, j] = pstarhatucb
                pstarhat_lcbs[i, j] = pstarhatlcb

                gaphat = fhatmeans[i] - pstarhatmean
                gaphats[i, j] = gaphat

                error = gaphat - gap
                print "Error between a gap estimator with [%d] reps. and [%d] samples and the true gap is [%f]." % (
                    num_replications,
                    num_samples,
                    error,
                )
                errors[i, j] = error  # np.square( )

        # Estimate the true optimal value (using the mean field approx.)
        print "Computing the mean field approximation."
        (pstarhat_mf, ignore, ignore, ignore) = self.unconstr_diag.est_optval(
            HowEstUnconstrOptVal.MEANFIELD, stoch_objf, None, None, None, None
        )

        # Plot num. Monte Carlo samples vs. num. replications vs. gap error (i.e., gaphat - gap)
        print "Plotting results to [" + fn1 + "] (on a linear/log scale)."
        print "FYI, the errors are:"
        print errors

        fig = plt.figure()
        plt.rc("text", usetex=True)
        ax = fig.gca(projection="3d")
        (num_reps_list_meshed, num_samps_list_meshed) = np.meshgrid(
            num_replications_list, num_samples_list
        )
        surf = ax.plot_surface(
            num_reps_list_meshed,
            num_samps_list_meshed,
            errors,
            rstride=1,
            cstride=1,
            cmap=cm.autumn,
            linewidth=0,
            antialiased=False,
            alpha=0.75,
        )
        # cont = plt.contour(num_reps_list_meshed, num_samps_list_meshed, errors)

        ax.set_xlabel(r"\# reps.")
        ax.set_ylabel(r"\# samples")
        ax.set_zlabel(r"error")  # $\hat{\Delta}_{M,N} - \Delta$
        ax.xaxis.set_scale("log")
        ax.yaxis.set_scale("log")
        plt.gca().invert_yaxis()
        plt.gca().invert_xaxis()
        fig.colorbar(surf, shrink=0.5, aspect=5)
        fig.savefig(fn1, bbox_inches="tight")
        print ""

        # Plot gaphat and gap vs. num. samples
        print "Plotting results to [" + fn2 + "] (on a linear/log scale)."
        print "FYI, the gaphats are:"
        print gaphats

        num_reps_fixed = 10
        num_samps_addl = 1000
        idx = 0

        stoch_objf = cvx.sum_squares(A * x - b)
        print "Averaging the opt. vals. of [%d] SAAs with [%d] Monte Carlo samples to the stoch. least squares problem; requesting t-based conf. ints." % (
            num_reps_fixed,
            num_samps_addl,
        )
        (
            pstarhatmean,
            pstarhatvar_ignore,
            pstarhatucb,
            pstarhatlcb,
        ) = self.unconstr_diag.est_optval(
            HowEstUnconstrOptVal.AVERAGE_SAAS,
            stoch_objf,
            num_reps_fixed,
            num_samps_addl,
            ConfInt.T,
            alpha,
        )

        gaphat_addl = fhatmeans[idx] - pstarhatmean
        print "Error between a gap estimator with [%d] reps. and [%d] samples and the true gap is [%f]." % (
            num_reps_fixed,
            num_samps_addl,
            gaphat_addl - gap,
        )

        gaphats_plot = np.append(gaphats[idx, :], gaphat_addl)
        num_samples_list.append(num_samps_addl)
        gaps = np.tile(gap, len(num_samples_list))

        fig, ax = plt.subplots()
        ax.plot(
            num_samples_list,
            gaphats_plot,
            "-",
            label=r"$\hat{\Delta}_{M,N}$",
            color="blue",
            lw=self.linewidth,
        )
        ax.plot(
            num_samples_list,
            gaps,
            "-",
            label=r"$\Delta$",
            color="crimson",
            lw=self.linewidth,
        )

        ax.legend(loc="upper right", fontsize=self.fontsize)
        ax.set_xlabel(
            r"\# samples $N$ (\# replications $M$ = %d fixed)" % num_reps_fixed,
            fontsize=self.fontsize,
        )
        ax.set_xscale("log")
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontsize(self.fontsize_axis)
        fig.savefig(fn2, bbox_inches="tight")
        print ""

        # Plot phatstar vs. num. samples
        print "Plotting results to [" + fn3 + "] (on a linear/log scale)."

        pstarhats_plot = np.append(pstarhats[idx, :], pstarhatmean)
        pstarhat_ucbs_plot = np.append(pstarhat_ucbs[idx, :], pstarhatucb)
        pstarhat_lcbs_plot = np.append(pstarhat_lcbs[idx, :], pstarhatlcb)

        pstars = np.tile(pstar, len(num_samples_list))

        pstarhats_mf = np.tile(pstarhat_mf, len(num_samples_list))

        fig, ax = plt.subplots()
        plt.rc("text", usetex=True)
        ax.plot(
            num_samples_list,
            pstarhats_plot,
            "-",
            label=r"$\hat{p}^*_{M,N}$",
            color="blue",
            lw=self.linewidth,
        )
        ax.fill_between(
            num_samples_list,
            pstarhat_ucbs_plot,
            pstarhat_lcbs_plot,
            facecolor="lightskyblue",
            alpha=0.5,
        )
        ax.plot(
            num_samples_list,
            pstarhats_mf,
            "-",
            label=r"$\hat{p}^*_{\mathrm{mf}}$",
            color="forestgreen",
            lw=self.linewidth,
        )
        ax.plot(
            num_samples_list,
            pstars,
            "-",
            label=r"$p^*$",
            color="crimson",
            lw=self.linewidth,
        )

        ax.legend(loc="lower left", fontsize=self.fontsize)
        ax.set_xlabel(
            r"\# samples $N$ (\# replications $M$ = %d fixed)" % num_reps_fixed,
            fontsize=self.fontsize,
        )
        ax.set_xscale("log")
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin - 1, ymax)
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontsize(self.fontsize_axis)
        fig.savefig(fn3, bbox_inches="tight")
        print ""

        # Plot pstarhat vs. num. replications
        print "Plotting results to [" + fn4 + "] (on a linear/log scale)."

        num_reps_addl = 1000
        num_samps_fixed = 100
        idx = 1

        stoch_objf = cvx.sum_squares(A * x - b)
        print "Averaging the opt. vals. of [%d] SAAs with [%d] Monte Carlo samples to the stoch. least squares problem; requesting t-based conf. ints." % (
            num_reps_addl,
            num_samps_fixed,
        )
        (
            pstarhatmean,
            pstarhatvar_ignore,
            pstarhatucb,
            pstarhatlcb,
        ) = self.unconstr_diag.est_optval(
            HowEstUnconstrOptVal.AVERAGE_SAAS,
            stoch_objf,
            num_reps_addl,
            num_samps_fixed,
            ConfInt.T,
            alpha,
        )

        pstarhats_plot = np.append(pstarhats[:, idx], pstarhatmean)
        pstarhat_ucbs_plot = np.append(pstarhat_ucbs[:, idx], pstarhatucb)
        pstarhat_lcbs_plot = np.append(pstarhat_lcbs[:, idx], pstarhatlcb)

        num_replications_list.append(num_reps_addl)

        fig, ax = plt.subplots()
        plt.rc("text", usetex=True)

        ax.plot(
            num_replications_list,
            pstarhats_plot,
            "-",
            label=r"$\hat{p}^*_{M,N}$",
            color="blue",
            lw=self.linewidth,
        )
        ax.fill_between(
            num_replications_list,
            pstarhat_ucbs_plot,
            pstarhat_lcbs_plot,
            facecolor="lightskyblue",
            alpha=0.5,
        )

        # ax.plot(num_replications_list, pstarhatmeans_boot, "-", label=r"$\hat{p}^*_{M,N,\textrm{boot}}$", color="darkorchid", lw=self.linewidth)
        # ax.fill_between(num_replications_list, pstarhatucbs_boot, pstarhatlcbs_boot, facecolor="plum", alpha=0.25)

        ax.plot(
            num_replications_list,
            pstarhats_mf,
            "-",
            label=r"$\hat{p}^*_{\mathrm{mf}}$",
            color="forestgreen",
            lw=self.linewidth,
        )

        ax.plot(
            num_replications_list,
            pstars,
            "-",
            label=r"$p^*$",
            color="crimson",
            lw=self.linewidth,
        )

        ax.legend(loc="lower left", fontsize=self.fontsize)
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin - 1, ymax)
        ax.set_xlabel(
            r"\# replications $M$ (\# samples $N$ = %d fixed)" % num_samps_fixed,
            fontsize=self.fontsize,
        )
        ax.set_xscale("log")
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontsize(self.fontsize_axis)
        fig.savefig(fn4, bbox_inches="tight")
        print ""

        # Plot gaphat and gap vs. num. replications
        print "Plotting results to [" + fn5 + "] (on a linear/log scale)."

        num_samps_fixed = 100
        idx = 1
        gaphat_addl = fhatmeans[2] - pstarhatmean
        gaphats_plot = np.append(gaphats[:, idx], gaphat_addl)

        fig, ax = plt.subplots()
        ax.plot(
            num_replications_list,
            gaphats_plot,
            "-",
            label=r"$\hat{\Delta}_{M,N}$",
            color="blue",
            lw=self.linewidth,
        )
        ax.plot(
            num_replications_list,
            gaps,
            "-",
            label=r"$\Delta$",
            color="crimson",
            lw=self.linewidth,
        )

        ax.legend(loc="upper right", fontsize=self.fontsize)
        ax.set_xlabel(
            r"\# replications $M$ (\# samples $N$ = %d fixed)" % num_samps_fixed,
            fontsize=self.fontsize,
        )
        ax.set_xscale("log")
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontsize(self.fontsize_axis)
        fig.savefig(fn5, bbox_inches="tight")
        print "All done."

    def test_least_squares(self, m, n, want_plot_saa_quality=False):
        # Create problem data
        Aij_mu, Aij_cov = 1, 1
        bi_mu, bi_cov = -1, 1
        A = cvx.stochastic.RandomVariableFactory().create_normal_rv(
            Aij_mu, Aij_cov, (m, n)
        )
        b = cvx.stochastic.RandomVariableFactory().create_normal_rv(
            bi_mu, bi_cov, (m, 1)
        )

        P = np.tile(m * Aij_mu * Aij_mu, (n, n))
        np.fill_diagonal(P, m * (math.pow(Aij_mu, 2) + Aij_cov))
        q = np.tile(m * Aij_mu * bi_mu, n)
        r = m * (math.pow(bi_mu, 2) + bi_cov)

        alpha = 0.01

        # Plot the optimal value of the sample average approximation vs. num. samples used to construct the SAA
        if want_plot_saa_quality:
            num_samples_list = [
                1,
                10,
                100,
                1000,
            ]  # Cut (too costly): ,10000].  Not used for any other tests.
            num_replications = 10  # Not used for any other tests
            alpha = 0.05  # Not used for any other tests
            fn = "fig1.png"

            self.test_least_squares_plot_saa_quality(
                A, b, P, q, r, num_samples_list, fn, num_replications, alpha
            )
            print ""

            return

        # Plot an estimate of the true objective value at a point vs. num. replications used to construct the estimate
        num_replications_list = [10, 100, 1000]  # Cut (too costly): ,10000]
        fn = "fig2.png"

        x = cvx.Variable(n)  # Get a candidate solution x first
        num_samples = 1
        prob = cvx.Problem(
            cvx.Minimize(
                cvx.stochastic.expectation(cvx.sum_squares(A * x - b), num_samples)
            )
        )
        prob.solve()

        (fhatmeans, fhatvars, fstars) = self.test_least_squares_plot_cost_ests(
            A, b, P, q, r, x, num_replications_list, alpha, fn
        )
        print ""

        # Make lots of other plots
        num_replications_list = [10, 100]
        num_samples_list = [10, 100]
        fn1, fn2, fn3, fn4, fn5 = (
            "fig3.png",
            "fig4.png",
            "fig5.png",
            "fig6.png",
            "fig12.png",
        )

        self.test_least_squares_plot_optval_ests(
            A,
            b,
            P,
            q,
            r,
            num_replications_list,
            num_samples_list,
            alpha,
            fn1,
            fn2,
            fn3,
            fn4,
            fn5,
            fhatmeans,
            fstars,
        )  # Cut (no longer used): (pstarhatmeans, pstarhatvars, pstars) =
        print ""

    def test_QCQP_compute_fhat(self, q0, r0, x, num_replications, alpha):
        n = q0.size

        fhats = []
        for i in range(num_replications):
            A0 = np.random.randn(n, n)
            P0 = A0.T.dot(A0)
            fhat = (0.5 * cvx.quad_form(x, P0) + x.T.dot(q0) + r0).value
            fhats.append(fhat)

        fhat_avg = np.mean(fhats)

        fhat_se = scipy.stats.sem(fhats)
        critval = UnconstrDiagnostics.get_critval(alpha)
        fhat_ucb = fhat_avg + critval * fhat_se
        fhat_lcb = fhat_avg - critval * fhat_se

        return (fhat_avg, fhat_ucb, fhat_lcb)

    def test_QCQP_compute_pstarhat(
        self, q0, r0, q1, r1, x, num_replications, num_samples, alpha
    ):
        n = q0.size

        pstarhats = []
        for k in range(num_replications):
            obj, constr = 0, 0
            for l in range(num_samples):
                A0 = np.random.randn(n, n)
                P0 = A0.T.dot(A0)
                obj += (1.0 / num_samples) * (
                    0.5 * cvx.quad_form(x, P0) + x.T * q0 + r0
                )

                A1 = np.random.randn(n, n)
                P1 = A1.T.dot(A1)
                constr += (1.0 / num_samples) * (
                    0.5 * cvx.quad_form(x, P1) + x.T * q1 + r1
                )

            pstarhat = self.constr_diag.est_optval(
                HowEstConstrOptVal.LAGRANGIAN, obj, [constr <= 0]
            )
            pstarhats.append(pstarhat)

        pstarhat_avg = np.mean(pstarhats)

        pstarhat_se = scipy.stats.sem(pstarhats)
        critval = UnconstrDiagnostics.get_critval(
            alpha, HowGetCritVal.T, num_replications - 1
        )
        pstarhat_ucb = pstarhat_avg + critval * pstarhat_se
        pstarhat_lcb = pstarhat_avg - critval * pstarhat_se

        return (pstarhat_avg, pstarhat_ucb, pstarhat_lcb)

    def test_QCQP(
        self,
        n=5,
        num_samples_list=[10, 100],
        num_replications_list=[10, 100],
        alpha=0.05,
        fn1="fig9.png",
        fn2="fig10.png",
        fn3="fig11.png",
    ):  # Cut (too costly): ,1000
        # Create problem data
        np.random.seed(
            4
        )  # I had to play with the seed to get the QCQP constraint in the true problem below to be feasible (similar to http://web.cvxr.com/cvx/examples/cvxbook/Ch05_duality/html/qcqp.html)

        Aij_mu, Aij_cov = 0, 1
        q0 = np.random.randn(n)
        q1 = np.random.randn(n)
        r0 = np.random.randn()
        r1 = np.random.randn()

        # Compute the true optimal value
        P = np.tile(n * Aij_mu * Aij_mu, (n, n))
        np.fill_diagonal(P, n * (math.pow(Aij_mu, 2) + Aij_cov))

        x = cvx.Variable(n)
        prob = cvx.Problem(
            cvx.Minimize(0.5 * cvx.quad_form(x, P) + x.T * q0 + r0),
            [0.5 * cvx.quad_form(x, P) + x.T * q1 + r1 <= 0],
        )
        prob.solve()
        pstar = prob.value

        # Compute the true cost
        xtilde = cvx.Variable(n)
        prob = cvx.Problem(
            cvx.Minimize(0), [0.5 * cvx.quad_form(xtilde, P) + x.T * q1 + r1 <= 0]
        )  # Just get a feasible point
        prob.solve()

        xtilde = xtilde.value
        fval = (0.5 * cvx.quad_form(xtilde, P) + xtilde.T.dot(q0) + r0).value

        # Compute the true optimality gap
        gap = fval - pstar

        # Estimate the optimality gap
        X = len(num_replications_list)
        Y = len(num_samples_list)

        pstarhats = np.zeros((X, Y))
        pstarhat_ucbs = np.zeros((X, Y))
        pstarhat_lcbs = np.zeros((X, Y))

        fhats = np.zeros(X)

        gaphats = np.zeros((X, Y))

        errors = np.zeros((X, Y))
        for i in range(X):  # range(X-1,-1,-1):
            num_replications = num_replications_list[i]

            # Estimate the true cost
            (fhat, fhat_ucb, fhat_lcb) = self.test_QCQP_compute_fhat(
                q0, r0, xtilde, num_replications, alpha
            )
            fhats[i] = fhat

            for j in range(Y):  # range(Y-1,-1,-1):
                num_samples = num_samples_list[j]

                # Estimate the true optimal value
                print "Minimizing Lagrangian [%d] times with [%d] Monte Carlo samples." % (
                    num_replications,
                    num_samples,
                )
                (
                    pstarhat,
                    pstarhat_ucb,
                    pstarhat_lcb,
                ) = self.test_QCQP_compute_pstarhat(
                    q0, r0, q1, r1, x, num_replications, num_samples, alpha
                )
                pstarhats[i, j] = pstarhat
                pstarhat_ucbs[i, j] = pstarhat_ucb
                pstarhat_lcbs[i, j] = pstarhat_lcb

                gaphat = fhat - pstarhat
                gaphats[i, j] = gaphat

                error = gaphat - gap
                print "Error between a gap estimator with [%d] reps. and [%d] samples and the true gap is [%f]." % (
                    num_replications,
                    num_samples,
                    error,
                )
                errors[i, j] = error  # np.square( )

        # Plot num. Monte Carlo samples vs. num. replications vs. gap error (i.e., gaphat - gap)
        print "Plotting results to [" + fn1 + "] (on a linear/log scale)."
        print "FYI, the errors are:"
        print errors

        fig = plt.figure()
        plt.rc("text", usetex=True)
        ax = fig.gca(projection="3d")
        (num_reps_list_meshed, num_samps_list_meshed) = np.meshgrid(
            num_replications_list, num_samples_list
        )
        surf = ax.plot_surface(
            num_reps_list_meshed,
            num_samps_list_meshed,
            errors,
            rstride=1,
            cstride=1,
            cmap=cm.autumn,
            linewidth=0,
            antialiased=False,
            alpha=0.75,
        )
        # cont = plt.contour(num_reps_list_meshed, num_samps_list_meshed, errors)

        ax.set_xlabel(r"\# reps.")
        ax.set_ylabel(r"\# samples")
        ax.set_zlabel(r"error")  # $\hat{\Delta}_{M,N} - \Delta$
        ax.xaxis.set_scale("log")
        ax.yaxis.set_scale("log")
        # plt.gca().invert_yaxis()
        # plt.gca().invert_xaxis()
        fig.colorbar(surf, shrink=0.5, aspect=5)
        fig.savefig(fn1, bbox_inches="tight")

        # Plot gaphat and gap vs. num. samples
        print "Plotting results to [" + fn2 + "] (on a linear/log scale)."
        print "FYI, the gaphats are:"
        print gaphats

        num_reps_fixed = 10  # Maybe unnecessary and too costly: 100
        num_samps_addl = 1000
        idx = 0

        print "Minimizing Lagrangian [%d] times with [%d] Monte Carlo samples." % (
            num_reps_fixed,
            num_samps_addl,
        )
        (
            pstarhat_addl,
            pstarhat_addl_ucb,
            pstarhat_addl_lcb,
        ) = self.test_QCQP_compute_pstarhat(
            q0, r0, q1, r1, x, num_reps_fixed, num_samps_addl, alpha
        )

        gaphat_addl = fhats[idx] - pstarhat_addl
        print "Error between a gap estimator with [%d] reps. and [%d] samples and the true gap is [%f]." % (
            num_reps_fixed,
            num_samps_addl,
            gaphat_addl - gap,
        )

        gaphats_plot = np.append(gaphats[idx, :], gaphat_addl)
        num_samples_list.append(num_samps_addl)
        gaps = np.tile(gap, len(num_samples_list))

        fig, ax = plt.subplots()
        ax.plot(
            num_samples_list,
            gaphats_plot,
            "-",
            label=r"$\hat{\Delta}_{M,N}$",
            color="blue",
            lw=self.linewidth,
        )
        ax.plot(
            num_samples_list,
            gaps,
            "-",
            label=r"$\Delta$",
            color="crimson",
            lw=self.linewidth,
        )

        ax.legend(loc="upper right", fontsize=self.fontsize)
        ax.set_xlabel(
            r"\# samples $N$ (\# replications $M$ = %d fixed)" % num_reps_fixed,
            fontsize=self.fontsize,
        )
        ax.set_xscale("log")
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontsize(self.fontsize_axis)
        fig.savefig(fn2, bbox_inches="tight")

        # Plot phatstar vs. num. samples
        print "Plotting results to [" + fn3 + "] (on a linear/log scale)."

        pstarhats_plot = np.append(pstarhats[idx, :], pstarhat_addl)
        pstarhat_ucbs_plot = np.append(pstarhat_ucbs[idx, :], pstarhat_addl_ucb)
        pstarhat_lcbs_plot = np.append(pstarhat_lcbs[idx, :], pstarhat_addl_lcb)

        pstars = np.tile(pstar, len(num_samples_list))

        fig, ax = plt.subplots()
        plt.rc("text", usetex=True)
        ax.plot(
            num_samples_list,
            pstarhats_plot,
            "-",
            label=r"$\hat{p}^*_{M,N}$",
            color="blue",
            lw=self.linewidth,
        )
        ax.fill_between(
            num_samples_list,
            pstarhat_ucbs_plot,
            pstarhat_lcbs_plot,
            facecolor="lightskyblue",
            alpha=0.5,
        )
        ax.plot(
            num_samples_list,
            pstars,
            "-",
            label=r"$p^*$",
            color="crimson",
            lw=self.linewidth,
        )

        ax.legend(loc="lower right", fontsize=self.fontsize)
        ax.set_xlabel(
            r"\# samples $N$ (\# replications $M$ = %d fixed)" % num_reps_fixed,
            fontsize=self.fontsize,
        )
        ax.set_xscale("log")
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontsize(self.fontsize_axis)
        fig.savefig(fn3, bbox_inches="tight")
        print "All done."

    def test_port_opt_plot_saa_quality(
        self,
        n=50,
        num_samples_list=[1, 10, 100, 1000],
        fn_objvals="fig7.png",
        fn_probs="fig8.png",
        num_replications=10,
        alpha=0.05,
    ):  # Cut (too costly): ,10000
        # Create problem data
        pbar = 10 + np.random.randn(n)
        A = np.random.rand(n, n)
        Sigma = 5 * A.T.dot(A)
        beta = 0.05  # Probability of portfolio loss (should be small)
        o = np.ones((n, 1))

        # Solve the SAA
        pstarhat_avgs, pstarhat_ucbs, pstarhat_lcbs = [], [], []
        prob_avgs, prob_ucbs, prob_lcbs = [], [], []
        critval = UnconstrDiagnostics.get_critval(
            alpha, HowGetCritVal.T, num_replications - 1
        )
        for num_samples in num_samples_list:
            print "Solving [%d] times a SAA to the stochastic port. opt. problem with [%d] Monte Carlo samples." % (
                num_replications,
                num_samples,
            )
            pstarhats = []
            probs = []
            for i in range(num_replications):
                x = cvx.Variable(n)
                p = cvx.stochastic.RandomVariableFactory().create_normal_rv(pbar, Sigma)
                cc = cvx.stochastic.prob(-x.T * p >= 0, num_samples)
                prob = cvx.Problem(
                    cvx.Maximize(x.T * pbar), [cc <= beta, x.T * o == 1, x >= -0.1]
                )
                prob.solve()  # solver=CVXOPT
                pstarhats.append(prob.value)
                probs.append(cc.get_prob())

            pstarhat_avg = np.mean(pstarhats)
            pstarhat_avgs.append(pstarhat_avg)

            pstarhat_se = scipy.stats.sem(pstarhats)
            pstarhat_ucbs.append(pstarhat_avg + critval * pstarhat_se)
            pstarhat_lcbs.append(pstarhat_avg - critval * pstarhat_se)

            prob_avg = np.mean(probs)
            prob_avgs.append(prob_avg)

            prob_se = scipy.stats.sem(probs)
            prob_ucbs.append(prob_avg + critval * prob_se)
            prob_lcbs.append(prob_avg - critval * prob_se)

        # Solve the true problem
        var = x.T * pbar >= scipy.stats.norm.ppf(1 - beta) * cvx.norm2(
            scipy.linalg.sqrtm(Sigma) * x
        )
        prob = cvx.Problem(cvx.Maximize(x.T * pbar), [var, x.T * o == 1, x >= -0.1])
        print "Solving the true stochastic port. opt. problem."
        prob.solve()
        pstars_var = np.tile(prob.value, len(num_samples_list))

        # Solve the true problem (with an exact CVaR constraint)
        cvar = x.T * pbar >= np.exp(
            -np.power(scipy.stats.norm.ppf(1 - beta), 2) / 2
        ) / (np.sqrt(2 * np.pi) * beta) * cvx.norm2(scipy.linalg.sqrtm(Sigma) * x)
        prob = cvx.Problem(cvx.Maximize(x.T * pbar), [cvar, x.T * o == 1, x >= -0.1])
        print "Solving the true stochastic port. opt. problem (with an exact CVaR constraint)."
        prob.solve()
        pstars_cvar = np.tile(prob.value, len(num_samples_list))

        # Make plots
        print "Plotting results to [" + fn_objvals + "] (on a linear/log scale)."
        fig, ax1 = plt.subplots()
        plt.rc("text", usetex=True)
        ax1.plot(
            num_samples_list,
            pstars_var,
            "-",
            label=r"$p^*$ (VaR)",
            color="forestgreen",
            lw=self.linewidth,
        )

        ax1.plot(
            num_samples_list,
            pstars_cvar,
            "-",
            label=r"$p^*$ (CVaR)",
            color="crimson",
            lw=self.linewidth,
        )

        ax1.plot(
            num_samples_list,
            pstarhat_avgs,
            "-",
            label=r"$\hat{p}^*_N$ (CVaR)",
            color="blue",
            lw=self.linewidth,
        )
        ax1.fill_between(
            num_samples_list,
            pstarhat_ucbs,
            pstarhat_lcbs,
            facecolor="lightskyblue",
            alpha=0.5,
        )
        # ax1.set_ylabel("objective value", fontsize=self.fontsize)

        ax1.legend(loc="lower left", fontsize=self.fontsize)

        ax1.set_xlabel(r"\# Monte Carlo samples $N$", fontsize=self.fontsize)
        ax1.set_xscale("log")

        # ax2 = ax1.twinx()

        for label in (
            ax1.get_xticklabels() + ax1.get_yticklabels()
        ):  # See http://stackoverflow.com/questions/3899980/how-to-change-the-font-size-on-a-matplotlib-plot
            label.set_fontsize(self.fontsize_axis)
        fig.savefig(fn_objvals, bbox_inches="tight")

        print "Plotting results to [" + fn_probs + "] (on a linear/log scale)."
        fig, ax2 = plt.subplots()
        plt.rc("text", usetex=True)
        ax2.plot(
            num_samples_list,
            prob_avgs,
            "-",
            label=r"$\mathop{\bf E{}} (-x^T \bar{p} / \alpha + 1)_+$",
            color="blue",
            lw=self.linewidth,
        )
        ax2.fill_between(
            num_samples_list, prob_ucbs, prob_lcbs, facecolor="lightskyblue", alpha=0.5
        )
        ax2.plot(
            num_samples_list,
            np.tile(beta, len(num_samples_list)),
            "-",
            label=r"$\beta$",
            color="crimson",
            lw=self.linewidth,
        )

        # ax2.set_ylabel("probability", fontsize=self.fontsize)
        # ax2.set_ylim((0,beta+0.01))
        ax2.legend(loc="lower right", fontsize=self.fontsize)
        ax2.set_xlabel(r"\# Monte Carlo samples $N$", fontsize=self.fontsize)
        ax2.set_xscale("log")
        for label in ax2.get_xticklabels() + ax2.get_yticklabels():
            label.set_fontsize(self.fontsize_axis)
        fig.savefig(fn_probs, bbox_inches="tight")
        print "All done."


if __name__ == "__main__":

    test_diag = TestDiagnostics()

    # Note: all the methods below work, I just commented some of them out so I could efficiently test them 1-by-1
    # test_diag.test_least_squares(m=100, n=50, want_plot_saa_quality=True)
    test_diag.test_least_squares(m=10, n=5)

    # test_diag.test_port_opt_plot_saa_quality()

    # test_diag.test_QCQP()



from cvxstoc.expectation import expectation
from cvxstoc.random_variable import (
    RandomVariable,
    RandomVariableFactory,
    NormalRandomVariable,
    CategoricalRandomVariable,
    UniformRandomVariable,
)
from cvxstoc.prob import prob, Phi



import copy, queue
import numpy

from cvxpy.transforms.partial_optimize import PartialProblem

from cvxstoc.random_variable import RandomVariable
from cvxstoc.utils import replace_rand_vars


def expectation(
    expr, num_samples=0, want_de=False, want_mf=False, num_burnin_samples=0
):
    if want_de:
        return construct_det_equiv(expr)
    elif want_mf:
        return construct_mean_field(expr)
    else:
        return construct_saa(expr, num_samples, num_burnin_samples)


def get_all_poss_combs_of_rv_realizations(rvs):
    num_rvs = len(rvs)
    num_combs = 1
    for rv in rvs:
        num_combs *= len(rv._metadata["vals"])

    idx_combs = numpy.zeros((num_combs, num_rvs))

    for i, rv in enumerate(rvs):

        num_reps = 1
        if i < (num_rvs - 1):
            for j in range(i + 1, num_rvs):
                num_reps *= len(rvs[j]._metadata["vals"])

        num_tiles = 1
        if i > 0:
            for j in range(i):
                num_tiles *= len(rvs[j]._metadata["vals"])

        idxes = range(len(rv._metadata["vals"]))
        col = numpy.tile(numpy.repeat(idxes, num_reps), num_tiles)

        idx_combs[:, i] = col[:]

    return idx_combs


def construct_mean_field(expr):
    rvs2samples, rvs = clamp_or_sample_rvs(
        expr,
        rvs2samples={},
        want_de=False,
        want_mf=True,
        num_samples=0,
        num_burnin_samples=0,
    )

    return clamp_or_sample_rvs(
        replace_rand_vars(expr),
        rvs2samples,
        want_de=False,
        want_mf=True,
        num_samples=0,
        num_burnin_samples=0,
    )


def construct_det_equiv(expr):
    rvs2samples, rvs = clamp_or_sample_rvs(
        expr,
        rvs2samples={},
        want_de=True,
        want_mf=False,
        num_samples=0,
        num_burnin_samples=0,
    )

    if not rvs2samples.keys():
        return expr

    else:
        combs = get_all_poss_combs_of_rv_realizations(rvs)
        num_combs = combs.shape[0]

        mul_exprs = []
        for s in range(num_combs):

            expr_copy = replace_rand_vars(expr)
            expr_copy_realized = clamp_or_sample_rvs(
                expr_copy,
                rvs2samples,
                want_de=True,
                want_mf=False,
                sample_idxes=combs[s, :],
            )

            prob = 1
            for i, rv in enumerate(rvs):
                idx = int(combs[s, i])
                prob *= rv._metadata["probs"][idx]

            mul_exprs.append(prob * expr_copy_realized)

        return sum(mul_exprs)


def construct_saa(expr, num_samples, num_burnin_samples):
    # Get samples of each RandomVariable in expr.
    rvs2samples, ignore = clamp_or_sample_rvs(
        expr,
        rvs2samples={},
        want_de=False,
        want_mf=False,
        num_samples=num_samples,
        num_burnin_samples=num_burnin_samples,
    )

    # If expr contained no RandomVariables, we're done.
    if not rvs2samples.keys():
        return expr

    # Else expr contained RandomVariables.
    else:
        mul_exprs = []
        for s in range(num_samples - num_burnin_samples):

            expr_copy = replace_rand_vars(expr)  # Deep copy expr (num_samples times)
            expr_copy_realized = clamp_or_sample_rvs(
                expr_copy, rvs2samples, want_de=False, want_mf=False, sample_idx=s
            )
            # Plug in a realization of each RandomVariable
            prob = 1.0 / (num_samples - num_burnin_samples)
            mul_exprs.append(prob * expr_copy_realized)

        # Return an AddExpression using all the realized deep copies to the caller.
        return sum(mul_exprs)


def clamp_or_sample_rvs(
    expr,
    rvs2samples={},
    want_de=None,
    want_mf=None,
    num_samples=None,
    num_burnin_samples=None,
    sample_idx=None,
    sample_idxes=None,
):
    # Walk expr and "process" each RandomVariable.
    if not rvs2samples.keys():
        draw_samples = True
    else:
        draw_samples = False

    my_queue = queue.Queue()
    my_queue.put(ExpectationqueueItem(expr))

    rvs = []
    rv_ctr = 0
    while True:
        queue_item = my_queue.get()
        cur_expr = queue_item._item
        if isinstance(cur_expr, RandomVariable):

            if draw_samples:
                rvs += [cur_expr]

                if cur_expr not in rvs2samples:
                    samples = cur_expr.sample(num_samples, num_burnin_samples)
                    rvs2samples[cur_expr] = samples

            else:
                if want_de:
                    idx = int(sample_idxes[rv_ctr])
                    cur_expr.value = cur_expr._metadata["vals"][idx]
                    rv_ctr += 1
                elif want_mf:
                    cur_expr.value = cur_expr.mean
                else:
                    cur_expr.value = rvs2samples[cur_expr][sample_idx]

        else:
            if isinstance(cur_expr, PartialProblem):
                my_queue.put(ExpectationqueueItem(cur_expr.args[0].objective))

                for constr in cur_expr.args[0].constraints:
                    my_queue.put(ExpectationqueueItem(constr))

            else:
                for arg in cur_expr.args:
                    my_queue.put(ExpectationqueueItem(arg))

        if my_queue.empty():
            break

    if draw_samples:
        return rvs2samples, rvs
    else:
        return expr


class ExpectationqueueItem:
    def __init__(self, item):
        self._item = item



from cvxstoc.random_variable import RandomVariable
from cvxpy.transforms.partial_optimize import PartialProblem
from cvxpy import Problem
import copy


def replace_rand_vars(obj):
    """Replaces all random variables with copies.

    Parameters
    ----------
    obj : Object
        The object to replace random variables in.

    Returns
    -------
    Object
        An object identical to obj, but with the random variables replaced.
    """
    if isinstance(obj, RandomVariable):
        return copy.deepcopy(obj)
    # Leaves other than random variables are preserved.
    elif len(obj.args) == 0:
        return obj
    elif isinstance(obj, PartialProblem):
        prob = obj.args[0]
        new_obj = replace_rand_vars(prob.objective)
        new_constr = []
        for constr in prob.constraints:
            new_constr.append(replace_rand_vars(constr))
        new_args = [Problem(new_obj, new_constr)]
        return obj.copy(new_args)
    # Parent nodes are copied.
    else:
        new_args = []
        for arg in obj.args:
            new_args.append(replace_rand_vars(arg))
        return obj.copy(new_args)



from setuptools import setup

setup(
    name="cvxstoc",
    version="0.2.2",
    author="Alnur Ali",
    packages=["cvxstoc"],
    license="GPLv3",
    description="A domain-specific language for modeling stochastic convex optimization problems in Python.",
    install_requires=[
        "cvxpy >= 0.3.5",
        "pymc >= 2.3.4",
        "numpy >= 1.10",
        "scipy >= 0.16",
    ],
    use_2to3=True,
)



import copy

import pymc.Model, pymc.MCMC

import numpy
import scipy

import cvxstoc.strings
from cvxpy import utilities, interface
import cvxpy.expressions.constants.constant, cvxpy.lin_ops.lin_utils


class RandomVariableFactory:
    _id = 0

    def __init__(self):
        pass

    def create_dirichlet_rv(
        self, alpha
    ):  # Note: alpha here is in R_+^K <==> x is in R^K ~ Dir(alpha)
        rv_inner_name = self.get_rv_name()
        rv_outer_name = self.get_rv_name()

        rv_pymc = pymc.CompletedDirichlet(
            name=rv_outer_name, D=pymc.Dirichlet(name=rv_inner_name, theta=alpha)
        )

        metadata = {}
        metadata["mu"] = [alpha[i] / numpy.sum(alpha) for i in range(len(alpha))]

        metadata["alpha"] = alpha

        return RandomVariable(rv=rv_pymc, metadata=metadata)

    def create_unif_rv(self, lower=0, upper=1, cont=True, shape=1):
        rv_name = self.get_rv_name()

        rv_pymc = None
        if cont:
            rv_pymc = pymc.Uniform(name=rv_name, lower=lower, upper=upper, size=shape)
        else:
            rv_pymc = pymc.DiscreteUniform(
                name=rv_name, lower=lower, upper=upper, size=shape
            )

        metadata = {}
        metadata["mu"] = (1.0 / 2.0) * (upper + lower)

        metadata["lower"] = lower
        metadata["upper"] = upper

        return RandomVariable(rv=rv_pymc, metadata=metadata)

    def create_categorical_rv(self, vals=None, probs=None, shape=1):
        rv_name = self.get_rv_name()

        rv_pymc = pymc.Categorical(name=rv_name, p=list(probs), size=shape)

        metadata = {}
        metadata["mu"] = len(probs) * numpy.asarray(probs)

        val_map = {}
        if vals:  # True when vals != None && vals != []
            for val in range(0, len(probs)):
                val_map[val] = vals[val]
        metadata["vals"] = vals
        metadata["probs"] = probs

        return RandomVariable(rv=rv_pymc, val_map=val_map, metadata=metadata)

    def create_normal_rv(self, mu, cov, shape=1):
        rv_name = self.get_rv_name()

        rv_pymc = None
        metadata = {}
        if isinstance(mu, numpy.ndarray):
            rv_pymc = pymc.MvNormal(name=rv_name, mu=mu, tau=scipy.linalg.inv(cov))

            metadata["mu"] = mu
            metadata["cov"] = cov
        else:
            rv_pymc = pymc.Normal(name=rv_name, mu=mu, tau=1.0 / cov, size=shape)

            if shape == 1:
                metadata["mu"] = mu
                metadata["cov"] = cov
            else:
                metadata["mu"] = numpy.tile(mu, shape)
                metadata["cov"] = cov

        return RandomVariable(rv=rv_pymc, metadata=metadata)

    def get_rv_name(self):
        return "rv" + str(RandomVariableFactory.get_next_avail_id())

    @staticmethod
    def get_next_avail_id():
        RandomVariableFactory._id += 1
        return RandomVariableFactory._id


def NormalRandomVariable(mean, cov, shape=1):
    return RandomVariableFactory().create_normal_rv(mu=mean, cov=cov, shape=shape)


def CategoricalRandomVariable(vals, probs, shape=1):
    return RandomVariableFactory().create_categorical_rv(
        vals=vals, probs=probs, shape=shape
    )


def UniformRandomVariable(lower=0, upper=1, cont=True, shape=1):
    return RandomVariableFactory().create_unif_rv(
        lower=lower, upper=upper, cont=cont, shape=shape
    )


class RandomVariable(cvxpy.expressions.constants.parameter.Parameter):
    def __init__(
        self, rv=None, model=None, name=None, val_map=None, metadata=None
    ):  # model == pymc.Model object
        if name is not None:
            self._name = name

        self._metadata = metadata

        self._val_map = val_map

        self.set_rv_model_and_maybe_name(rv, model)

        self.set_shape()

        super(RandomVariable, self).__init__(self._shape, self._name)

    @property
    def mean(self):
        return self._metadata["mu"]

    def set_rv_model_and_maybe_name(self, rv, model):
        if rv is not None and model is None:
            self._rv = rv
            self._model = pymc.Model([self._rv])

            self._name = self._rv.__name__

        elif rv is not None and model is not None:
            self._rv = rv
            self._model = model

            self._name = self._rv.__name__

        elif rv is None and model is not None:

            self._model = model

            self._rv = None
            for pymc_variable in self._model.variables:
                if pymc_variable.__name__ == self._name:
                    if isinstance(pymc_variable, pymc.CompletedDirichlet):
                        if pymc_variable.parents["D"].observed:
                            continue
                        # Success.
                    elif hasattr(pymc_variable, "observed"):
                        if pymc_variable.observed:
                            continue
                        # Success.
                    # Failure.
                    else:
                        raise Exception(strings.UNSUPPORTED_PYMC_RV)

                    self._rv = pymc_variable
                    break

            if self._rv is None:
                raise Exception(strings.CANT_FIND_PYMC_RV_IN_PYMC_MODEL_OBJ)

        else:
            raise Exception(strings.DIDNT_PASS_EITHER_RV_OR_MODEL)

    def set_shape(self):
        shape = ()
        if self.has_val_map():
            val = list(self._val_map.values())[0]

            if isinstance(val, int) or isinstance(val, float):
                shape = ()

            elif isinstance(val, numpy.ndarray):

                numpy_shape = val.shape
                if len(numpy_shape) == 1:
                    shape = (numpy_shape[0],)
                elif len(numpy_shape) == 2:
                    shape = (numpy_shape[0], numpy_shape[1])
                else:
                    raise Exception(strings.BAD_RV_DIMS)

            else:
                raise Exception(strings.BAD_VAL_MAP)

        else:
            pymc_shape = ()
            if isinstance(self._rv, pymc.CompletedDirichlet):
                pymc_shape = self._rv.parents["D"].shape
            else:
                pymc_shape = self._rv.shape

            if len(pymc_shape) == 0:
                shape = ()
            elif len(pymc_shape) == 1:
                shape = (pymc_shape[0],)
            elif len(pymc_shape) == 2:
                shape = (pymc_shape[0], pymc_shape[1])
            else:
                raise Exception(strings.BAD_RV_DIMS)

        self._shape = shape

    def name(self):
        # Override.
        if self.value is None:
            return self._name
        else:
            return str(self.value)

    def __repr__(self):
        # Override.
        return "RandomVariable(%s, %s, %s)" % (self.curvature, self.sign, self.size)

    def __eq__(self, rv):
        # Override.
        return self._name == rv._name

    def __hash__(self):
        # Override.
        return hash(self._name)

    def __deepcopy__(self, memo):
        # Override.
        return self.__class__(
            rv=self._rv,
            model=self._model,
            name=self.name(),
            val_map=self._val_map,
            metadata=self._metadata,
        )

    def sample(self, num_samples, num_burnin_samples=0):
        if num_samples == 0:
            return [None]

        mcmc = pymc.MCMC(self._model)
        mcmc.sample(num_samples, num_burnin_samples, progress_bar=False)
        samples = mcmc.trace(self._name)[:]

        if not self.has_val_map():
            return samples
        else:
            samples_mapped = [self._val_map[sample[0]] for sample in samples]
            return samples_mapped

    def has_val_map(self):
        if self._val_map is not None and len(self._val_map.values()) > 0:
            return True
        else:
            return False
